@@ -26,22 +26,28 @@ module Data.Conduit.List
     , consume
     , sinkNull
       -- ** Monadic
+    , foldMapM
     , foldM
     , mapM_
       -- * Conduits
       -- ** Pure
     , map
     , mapMaybe
+    , mapFoldable
     , catMaybes
+    , concat
     , concatMap
     , concatMapAccum
+    , scanl
     , groupBy
     , isolate
     , filter
       -- ** Monadic
     , mapM
     , iterM
+    , scanlM
     , mapMaybeM
+    , mapFoldableM
     , concatMapM
     , concatMapAccumM
       -- * Misc
@@ -62,8 +68,9 @@ import Prelude
     , either
     )
 import Data.Monoid (Monoid, mempty, mappend)
+import qualified Data.Foldable as F
 import Data.Conduit
-import Control.Monad (when, (<=<))
+import Control.Monad (when, (<=<), liftM)
 import Control.Monad.Trans.Class (lift)
 
 -- | Generate a source from a seed value.
@@ -154,6 +161,17 @@ foldMap f =
   where
     combiner accum = mappend accum . f
 
+-- | A monoidal strict left fold in a Monad.
+--
+-- Since 1.0.8
+foldMapM :: (Monad m, Monoid b)
+        => (a -> m b)
+        -> Consumer a m b
+foldMapM f =
+    foldM combiner mempty
+  where
+    combiner accum = liftM (mappend accum) . f
+
 -- | Apply the action to all values in the stream.
 --
 -- Since 0.3.0
@@ -279,6 +297,13 @@ mapMaybeM f = awaitForever $ maybe (return ()) yield <=< lift . f
 catMaybes :: Monad m => Conduit (Maybe a) m a
 catMaybes = awaitForever $ maybe (return ()) yield
 
+-- | Generalization of 'catMaybes'. It puts all values from
+--   'F.Foldable' into stream.
+--
+-- Since 1.0.6
+concat :: (Monad m, F.Foldable f) => Conduit (f a) m a
+concat = awaitForever $ F.mapM_ yield
+
 -- | Apply a transformation to all values in a stream, concatenating the output
 -- values.
 --
@@ -297,31 +322,53 @@ concatMapM f = awaitForever $ sourceList <=< lift . f
 --
 -- Since 0.3.0
 concatMapAccum :: Monad m => (a -> accum -> (accum, [b])) -> accum -> Conduit a m b
-concatMapAccum f =
+concatMapAccum f x0 = scanl f x0 =$= concat
+
+-- | Analog of 'Prelude.scanl' for lists.
+--
+-- Since 1.0.6
+scanl :: Monad m => (a -> s -> (s,b)) -> s -> Conduit a m b
+scanl f =
     loop
   where
-    loop accum =
-        await >>= maybe (return ()) go
+    loop s = await >>= F.mapM_ go
       where
-        go a = do
-            let (accum', bs) = f a accum
-            Prelude.mapM_ yield bs
-            loop accum'
+        go a = case f a s of
+                 (s',b) -> yield b >> loop s'
+
+-- | Monadic scanl.
+--
+-- Since 1.0.6
+scanlM :: Monad m => (a -> s -> m (s,b)) -> s -> Conduit a m b
+scanlM f =
+    loop
+  where
+    loop s = await >>= F.mapM_ go
+      where
+        go a = do (s',b) <- lift $ f a s
+                  yield b >> loop s'
 
 -- | 'concatMapM' with an accumulator.
 --
 -- Since 0.3.0
 concatMapAccumM :: Monad m => (a -> accum -> m (accum, [b])) -> accum -> Conduit a m b
-concatMapAccumM f =
-    loop
-  where
-    loop accum = do
-        await >>= maybe (return ()) go
-      where
-        go a = do
-            (accum', bs) <- lift $ f a accum
-            Prelude.mapM_ yield bs
-            loop accum'
+concatMapAccumM f x0 = scanlM f x0 =$= concat
+
+
+-- | Generalization of 'mapMaybe' and 'concatMap'. It applies function
+-- to all values in a stream and send values inside resulting
+-- 'Foldable' downstream.
+--
+-- Since 1.0.6
+mapFoldable :: (Monad m, F.Foldable f) => (a -> f b) -> Conduit a m b
+mapFoldable f = awaitForever $ F.mapM_ yield . f
+
+-- | Monadic variant of 'mapFoldable'.
+--
+-- Since 1.0.6
+mapFoldableM :: (Monad m, F.Foldable f) => (a -> m (f b)) -> Conduit a m b
+mapFoldableM f = awaitForever $ F.mapM_ yield <=< lift . f
+
 
 -- | Consume all values from the stream and return as a list. Note that this
 -- will pull all values into memory. For a lazy variant, see
