@@ -12,7 +12,7 @@ import qualified Data.Conduit.Lazy as CLazy
 import qualified Data.Conduit.Binary as CB
 import qualified Data.Conduit.Text as CT
 import Data.Conduit (runResourceT)
-import Data.Maybe (fromMaybe)
+import Data.Maybe   (fromMaybe,catMaybes)
 import qualified Data.List as DL
 import Control.Monad.ST (runST)
 import Data.Monoid
@@ -31,14 +31,22 @@ import Control.Monad.Trans.Class (lift)
 import Control.Monad.Trans.Writer (execWriter, tell, runWriterT)
 import Control.Monad.Trans.State (evalStateT, get, put)
 import Control.Applicative (pure, (<$>), (<*>))
-import Data.Functor.Identity (runIdentity)
+import Data.Functor.Identity (Identity,runIdentity)
 import Control.Monad (forever)
 import Data.Void (Void)
 import qualified Control.Concurrent.MVar as M
+import Control.Monad.Error (catchError, throwError, Error)
 
 (@=?) :: (Eq a, Show a) => a -> a -> IO ()
 (@=?) = flip shouldBe
 
+-- Quickcheck property for testing equivalence of list processing
+-- functions and their conduit counterparts
+equivToList :: Eq b => ([a] -> [b]) -> CI.Conduit a Identity b -> [a] -> Bool
+equivToList f conduit xs =
+  f xs == runIdentity (CL.sourceList xs C.$$ conduit C.=$= CL.consume)
+
+
 main :: IO ()
 main = hspec $ do
     describe "data loss rules" $ do
@@ -61,6 +69,28 @@ main = hspec $ do
             x <- runResourceT $ CL.sourceList [1..10] C.$$ CL.filter even C.=$ CL.consume
             x `shouldBe` filter even [1..10 :: Int]
 
+    prop "concat" $ equivToList (concat :: [[Int]]->[Int]) CL.concat
+
+    describe "mapFoldable" $ do
+        prop "list" $
+            equivToList (concatMap (:[]) :: [Int]->[Int]) (CL.mapFoldable  (:[]))
+        let f x = if odd x then Just x else Nothing
+        prop "Maybe" $
+            equivToList (catMaybes . map f :: [Int]->[Int]) (CL.mapFoldable f)
+
+    prop "scanl" $ equivToList (tail . scanl (+) 0 :: [Int]->[Int]) (CL.scanl (\a s -> (a+s,a+s)) 0)
+
+    -- mapFoldableM and scanlM are fully polymorphic in type of monad
+    -- so it suffice to check only with Identity.
+    describe "mapFoldableM" $ do
+        prop "list" $
+            equivToList (concatMap (:[]) :: [Int]->[Int]) (CL.mapFoldableM (return . (:[])))
+        let f x = if odd x then Just x else Nothing
+        prop "Maybe" $
+            equivToList (catMaybes . map f :: [Int]->[Int]) (CL.mapFoldableM (return . f))
+
+    prop "scanl" $ equivToList (tail . scanl (+) 0 :: [Int]->[Int]) (CL.scanlM (\a s -> return (a+s,a+s)) 0)
+
     describe "ResourceT" $ do
         it "resourceForkIO" $ do
             counter <- I.newIORef 0
@@ -102,6 +132,15 @@ main = hspec $ do
             x <- CL.sourceList [[4],[2],[3],[1]] C.$$ CL.foldMap (++[(9 :: Int)])
             x `shouldBe` [4,9,2,9,3,9,1,9]
 
+    describe "foldMapM" $ do
+        it "sums 1..10" $ do
+            Sum x <- CL.sourceList [1..(10 :: Int)] C.$$ CL.foldMapM (return . Sum)
+            x `shouldBe` sum [1..10]
+
+        it "preserves order" $ do
+            x <- CL.sourceList [[4],[2],[3],[1]] C.$$ CL.foldMapM (return . (++[(9 :: Int)]))
+            x `shouldBe` [4,9,2,9,3,9,1,9]
+
     describe "unfold" $ do
         it "works" $ do
             let f 0 = Nothing
@@ -359,7 +398,7 @@ main = hspec $ do
             (a, b) `shouldBe` (Just 1, [1..10])
 
     describe "text" $ do
-        let go enc tenc cenc = do
+        let go enc tenc tdec cenc = do
                 prop (enc ++ " single chunk") $ \chars -> runST $ runExceptionT_ $ do
                     let tl = TL.pack chars
                         lbs = tenc tl
@@ -370,19 +409,32 @@ main = hspec $ do
                     let tl = TL.pack chars
                         lbs = tenc tl
                         src = mconcat $ map (CL.sourceList . return . S.singleton) $ L.unpack lbs
+                        
                     ts <- src C.$= CT.decode cenc C.$$ CL.consume
                     return $ TL.fromChunks ts == tl
+
+                -- Check whether raw bytes are decoded correctly, in
+                -- particular that Text decoding produces an error if
+                -- and only if Conduit does.
+                prop (enc ++ " raw bytes") $ \bytes ->
+                    let lbs = L.pack bytes
+                        src = CL.sourceList $ L.toChunks lbs
+                        etl = C.runException $ src C.$= CT.decode cenc C.$$ CL.consume
+                        tl' = tdec lbs
+                    in  case etl of
+                          (Left _) -> (return $! TL.toStrict tl') `shouldThrow` anyException
+                          (Right tl) -> TL.fromChunks tl `shouldBe` tl'
                 prop (enc ++ " encoding") $ \chars -> runIdentity $ runExceptionT_ $ do
                     let tss = map T.pack chars
                         lbs = tenc $ TL.fromChunks tss
                         src = mconcat $ map (CL.sourceList . return) tss
                     bss <- src C.$= CT.encode cenc C.$$ CL.consume
                     return $ L.fromChunks bss == lbs
-        go "utf8" TLE.encodeUtf8 CT.utf8
-        go "utf16_le" TLE.encodeUtf16LE CT.utf16_le
-        go "utf16_be" TLE.encodeUtf16BE CT.utf16_be
-        go "utf32_le" TLE.encodeUtf32LE CT.utf32_le
-        go "utf32_be" TLE.encodeUtf32BE CT.utf32_be
+        go "utf8" TLE.encodeUtf8 TLE.decodeUtf8 CT.utf8
+        go "utf16_le" TLE.encodeUtf16LE TLE.decodeUtf16LE CT.utf16_le
+        go "utf16_be" TLE.encodeUtf16BE TLE.decodeUtf16BE CT.utf16_be
+        go "utf32_le" TLE.encodeUtf32LE TLE.decodeUtf32LE CT.utf32_le
+        go "utf32_be" TLE.encodeUtf32BE TLE.decodeUtf32BE CT.utf32_be
 
     describe "text lines" $ do
         it "works across split lines" $
@@ -884,5 +936,122 @@ main = hspec $ do
                 lbs' `shouldBe` lbs
                 fromIntegral len `shouldBe` L.length lbs'
 
+    describe "mtl instances" $ do
+        it "ErrorT" $ do
+            let src = flip catchError (const $ C.yield 4) $ do
+                    lift $ return ()
+                    C.yield 1
+                    lift $ return ()
+                    C.yield 2
+                    lift $ return ()
+                    () <- throwError DummyError
+                    lift $ return ()
+                    C.yield 3
+                    lift $ return ()
+            (src C.$$ CL.consume) `shouldBe` Right [1, 2, 4 :: Int]
+
+    describe "finalizers" $ do
+        it "promptness" $ do
+            imsgs <- I.newIORef []
+            let add x = liftIO $ do
+                    msgs <- I.readIORef imsgs
+                    I.writeIORef imsgs $ msgs ++ [x]
+                src' = C.bracketP
+                    (add "acquire")
+                    (const $ add "release")
+                    (const $ C.addCleanup (const $ add "inside") (mapM_ C.yield [1..5]))
+                src = do
+                    src' C.$= CL.isolate 4
+                    add "computation"
+                sink = CL.mapM (\x -> add (show x) >> return x) C.=$ CL.consume
+
+            res <- C.runResourceT $ src C.$$ sink
+
+            msgs <- I.readIORef imsgs
+            -- FIXME this would be better msgs `shouldBe` words "acquire 1 2 3 4 inside release computation"
+            msgs `shouldBe` words "acquire 1 2 3 4 release inside computation"
+
+            res `shouldBe` [1..4 :: Int]
+
+        it "left associative" $ do
+            imsgs <- I.newIORef []
+            let add x = liftIO $ do
+                    msgs <- I.readIORef imsgs
+                    I.writeIORef imsgs $ msgs ++ [x]
+                p1 = C.bracketP (add "start1") (const $ add "stop1") (const $ add "inside1" >> C.yield ())
+                p2 = C.bracketP (add "start2") (const $ add "stop2") (const $ add "inside2" >> C.await >>= maybe (return ()) C.yield)
+                p3 = C.bracketP (add "start3") (const $ add "stop3") (const $ add "inside3" >> C.await)
+
+            res <- C.runResourceT $ (p1 C.$= p2) C.$$ p3
+            res `shouldBe` Just ()
+
+            msgs <- I.readIORef imsgs
+            msgs `shouldBe` words "start3 inside3 start2 inside2 start1 inside1 stop3 stop2 stop1"
+
+        it "right associative" $ do
+            imsgs <- I.newIORef []
+            let add x = liftIO $ do
+                    msgs <- I.readIORef imsgs
+                    I.writeIORef imsgs $ msgs ++ [x]
+                p1 = C.bracketP (add "start1") (const $ add "stop1") (const $ add "inside1" >> C.yield ())
+                p2 = C.bracketP (add "start2") (const $ add "stop2") (const $ add "inside2" >> C.await >>= maybe (return ()) C.yield)
+                p3 = C.bracketP (add "start3") (const $ add "stop3") (const $ add "inside3" >> C.await)
+
+            res <- C.runResourceT $ p1 C.$$ (p2 C.=$ p3)
+            res `shouldBe` Just ()
+
+            msgs <- I.readIORef imsgs
+            msgs `shouldBe` words "start3 inside3 start2 inside2 start1 inside1 stop3 stop2 stop1"
+
+        describe "dan burton's associative tests" $ do
+            let tellLn = tell . (++ "\n")
+                finallyP fin = CI.addCleanup (const fin)
+                printer = CI.awaitForever $ lift . tellLn . show
+                idMsg msg = finallyP (tellLn msg) CI.idP
+                takeP 0 = return ()
+                takeP n = CI.awaitE >>= \ex -> case ex of
+                  Left _u -> return ()
+                  Right i -> CI.yield i >> takeP (pred n)
+
+                testPipe p = execWriter $ runPipe $ printer <+< p <+< CI.sourceList ([1..] :: [Int])
+
+                p1 = takeP (1 :: Int)
+                p2 = idMsg "foo"
+                p3 = idMsg "bar"
+
+                (<+<) = (CI.<+<)
+                runPipe = CI.runPipe
+
+                test1L = testPipe $ (p1 <+< p2) <+< p3
+                test1R = testPipe $ p1 <+< (p2 <+< p3)
+
+                test2L = testPipe $ (p2 <+< p1) <+< p3
+                test2R = testPipe $ p2 <+< (p1 <+< p3)
+
+                test3L = testPipe $ (p2 <+< p3) <+< p1
+                test3R = testPipe $ p2 <+< (p3 <+< p1)
+
+                verify testL testR p1' p2' p3'
+                  | testL == testR = return () :: IO ()
+                  | otherwise = error $ unlines
+                    [ "FAILURE"
+                    , ""
+                    , "(" ++ p1' ++ " <+< " ++ p2' ++ ") <+< " ++ p3'
+                    , "------------------"
+                    , testL
+                    , ""
+                    , p1' ++ " <+< (" ++ p2' ++ " <+< " ++ p3' ++ ")"
+                    , "------------------"
+                    , testR
+                    ]
+
+            it "test1" $ verify test1L test1R "p1" "p2" "p3"
+            -- FIXME this is broken it "test2" $ verify test2L test2R "p2" "p1" "p3"
+            it "test3" $ verify test3L test3R "p2" "p3" "p1"
+
 it' :: String -> IO () -> Spec
 it' = it
+
+data DummyError = DummyError
+    deriving (Show, Eq)
+instance Error DummyError
